/*
   mInts: Monster Signed Integers for Java
   Copyright (c) 1997 Lou Grinzo
*/

#include <windows.h>
#include "mInt.h"

/*
  To change the size of a mInt representation, change
  the next 6 constants as needed.  A mInt must always
  be a multiple of 4 bytes, and at least 8 bytes.

  You also must change the method implementations, like
  Add(), which resort to assembly language, various
  constant mInts throughout this code, and mInt.java.
*/

#define bits_in_mint     256
#define bits_in_component 32
#define ints_in_mint       8
#define bytes_in_mint     32

/* This is the maximum number of ASCII chars needed
   to represent a 256-bit number in hex, NOT including
   room for a terminating null, but allowing for a blank
   after each byte except the last one */

#define max_hex_chars_in_mint 95

/* This is the maximum number of ASCII chars needed
   to represent a 256-bit signed number in decimal,
   including one character for a possible leading '-',
   but not including a null.  */

#define max_dec_chars_in_mint 78

/* End of size-dependent constants */

#define sign_bit 0x80000000
#define all_ones 0xffffffff
#define max_int  0x7fffffff

/* Possible last_error values.  These must agree with
   the constants in mInt.java! */
#define mInt_OK        0
#define mInt_overflow  1
#define mInt_underflow 2
#define mInt_divide_by_zero 3
#define mInt_invalid_char 4

/* Table of values used by LeadingZeroes.  Each
   value is the number of leading binary zeroes in
   a byte equal to that value's index. */
static unsigned char lz_tab[256] =
{
  8, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

jint mint_10[ints_in_mint] = { 0, 0, 0, 0, 0, 0, 0, 10 };

/* The fieldIDs, which we cache for performance reasons */
static jfieldID global_int_FID;
static jfieldID global_le_FID;

/* Local functions */

static jboolean UAdd(JNIEnv *env, jobject this, jint *io,
  jint *input2);
static jboolean USubtract(JNIEnv *env, jobject this,
  jint *io, jint *input2);
static jboolean UMultiply(JNIEnv *env, jobject this,
  jint *io, jint *input2);
static jboolean UDivide(JNIEnv *env, jobject this,
  jint *numer, jint *denom, jint *quotient,jint *remainder);
static jint     UComp(jint *x, jint *y);
static void     Negate(jint *x);
static void     UInt32ToInt(jint *x, unsigned int y);
static int      LeadingZeroes(jint *x);
static jboolean ShiftLeft(JNIEnv *env, jobject this, jint *x);
static void     ShiftRight(jint *x);
static void     ShiftLeftNoError(jint *x);
static void     ShiftLeftNNoError(jint *x, int shift_bits);
static void     ShiftLeft32NoError(jint *x);
static jboolean PosToNeg(JNIEnv *env, jobject this, jint *x);
static BOOL     IsZero(jint *x);
static void     mInt_SetLastError(JNIEnv *env, jobject this,
  int new_last_error);

/* Retrieve and store the fieldIDs for the mInt representation
   and the last_error members.  We do this for performance reasons,
   since the GetFieldID() call is quite expensive. */

JNIEXPORT void JNICALL Java_mInt_cacheFID(JNIEnv *env,
  jclass the_class)
{
  /* We have to use both the name and a type signature
     to look up a field's fid, e.g. "[I" means "array of
     integers". */
  global_int_FID = (*env)->GetFieldID(env,the_class,
    "the_int","[I");
  global_le_FID = (*env)->GetStaticFieldID(env,the_class,
    "last_error","I");
}

/* Addition */

JNIEXPORT jboolean JNICALL Java_mInt_Add(JNIEnv *env, jobject this,
  jobject other)
{
  int temp, sign_io, sign_input2;
  jboolean ret_val = JNI_TRUE;

  jintArray io_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *io = (*env)->GetIntArrayElements(env,io_array,0);

  jintArray input2_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *input2 = (*env)->GetIntArrayElements(env,input2_array,0);

  sign_input2 = input2[0] & sign_bit;
  sign_io = io[0] & sign_bit;

  temp = 0;

__asm
  {
    mov edi,io
    mov esi,input2

    mov eax,[esi+28]
    add [edi+28],eax

    mov eax,[esi+24]
    adc [edi+24],eax

    mov eax,[esi+20]
    adc [edi+20],eax

    mov eax,[esi+16]
    adc [edi+16],eax

    mov eax,[esi+12]
    adc [edi+12],eax

    mov eax,[esi+8]
    adc [edi+8],eax

    mov eax,[esi+4]
    adc [edi+4],eax

    mov eax,[esi]
    adc [edi],eax

    jnc no_carry

    mov eax,1
    mov temp,eax

  no_carry:
  }

  /* Signed addition can only overflow/underflow when the
     inputs had the same sign */
  if(sign_io == sign_input2)
  {
    if((sign_io != 0) && (!(io[0] & sign_bit)))
    {
      /* Inputs were neg and result was
         pos --> underflow */
      mInt_SetLastError(env,this,mInt_underflow);
      ret_val = JNI_FALSE;
    }
    else
      if((sign_io == 0) && (io[0] & sign_bit))
      {
        /* Inputs were pos and result was
           neg --> overflow */
        mInt_SetLastError(env,this,mInt_overflow);
        ret_val = JNI_FALSE;
      }
  }

  (*env)->ReleaseIntArrayElements(env,input2_array,input2,0);
  (*env)->ReleaseIntArrayElements(env,io_array,io,0);

  return ret_val;
}

/* Addition, but without using the cached fileID. */

JNIEXPORT jboolean JNICALL Java_mInt_testAdd(JNIEnv *env,
  jobject this, jobject other)
{
  int temp, sign_io, sign_input2;
  jboolean ret_val = JNI_TRUE;

  jclass io_class = (*env)->GetObjectClass(env,this);
  jfieldID io_fid = (*env)->GetFieldID(env,io_class,"the_int",
    "[I");
  jintArray io_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *io = (*env)->GetIntArrayElements(env,io_array,0);

  jclass input2_class = (*env)->GetObjectClass(env,other);
  jfieldID input2_fid = (*env)->GetFieldID(env,input2_class,
    "the_int","[I");
  jintArray input2_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *input2 = (*env)->GetIntArrayElements(env,input2_array,0);

  sign_input2 = input2[0] & sign_bit;
  sign_io =     io[0] & sign_bit;

  temp = 0;

__asm
  {
    mov edi,io
    mov esi,input2

    mov eax,[esi+28]
    add [edi+28],eax

    mov eax,[esi+24]
    adc [edi+24],eax

    mov eax,[esi+20]
    adc [edi+20],eax

    mov eax,[esi+16]
    adc [edi+16],eax

    mov eax,[esi+12]
    adc [edi+12],eax

    mov eax,[esi+8]
    adc [edi+8],eax

    mov eax,[esi+4]
    adc [edi+4],eax

    mov eax,[esi]
    adc [edi],eax

    jnc no_carry

    mov eax,1
    mov temp,eax

  no_carry:
  }

  if(sign_io == sign_input2)
  {
    if((sign_io != 0) && (!(io[0] & sign_bit)))
    {
      mInt_SetLastError(env,this,mInt_underflow);
      ret_val = JNI_FALSE;
    }
    else
      if((sign_io == 0) && (io[0] & sign_bit))
      {
        mInt_SetLastError(env,this,mInt_overflow);
        ret_val = JNI_FALSE;
      }
  }

  (*env)->ReleaseIntArrayElements(env,input2_array,input2,0);
  (*env)->ReleaseIntArrayElements(env,io_array,io,0);

  return ret_val;
}

/* Addition, but showing how to create a Java mInt
   object from within a native method and return it. */

JNIEXPORT jobject JNICALL Java_mInt_test2Add(JNIEnv *env,
  jobject this, jobject other)
{
  /* Get a handle to one of our constructors */
  jclass the_class = (*env)->GetObjectClass(env,this);
  jmethodID the_method = (*env)->GetMethodID(env,the_class,
    "<init>","(I)V");

  /* Get access to this' mInt */
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);

  /* Create a new mInt and copy this' value into it */
  jobject new_mInt = (*env)->NewObject(env,the_class,
    the_method,0);

  jintArray y_array = (*env)->GetObjectField(env,new_mInt,
    global_int_FID);
  jint *y = (*env)->GetIntArrayElements(env,y_array,0);

  memcpy(y,x,bytes_in_mint);

  /* Done with these so we can release them */
  (*env)->ReleaseIntArrayElements(env,y_array,y,0);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);

  /* Use the Add() implementation to do the real work */
  Java_mInt_Add(env,new_mInt,other);

  return new_mInt;
}

/* Subtraction */

JNIEXPORT jboolean JNICALL Java_mInt_Subtract(JNIEnv *env,
  jobject this, jobject other)
{
  int temp, sign_io, sign_input2;
  jboolean ret_val = JNI_TRUE;

  jintArray io_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *io = (*env)->GetIntArrayElements(env,io_array,0);

  jintArray input2_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *input2 = (*env)->GetIntArrayElements(env,input2_array,0);

  sign_input2 = input2[0] & sign_bit;
  sign_io =     io[0] & sign_bit;

  temp = 0;

__asm
  {
    mov edi,io
    mov esi,input2

    mov eax,[esi+28]
    sub [edi+28],eax

    mov eax,[esi+24]
    sbb [edi+24],eax

    mov eax,[esi+20]
    sbb [edi+20],eax

    mov eax,[esi+16]
    sbb [edi+16],eax

    mov eax,[esi+12]
    sbb [edi+12],eax

    mov eax,[esi+8]
    sbb [edi+8],eax

    mov eax,[esi+4]
    sbb [edi+4],eax

    mov eax,[esi]
    sbb [edi],eax

    jnc no_carry

    mov eax,1
    mov temp,eax

  no_carry:
  }

  /* Signed subtraction can only overflow/underflow when
     the inputs had different signs */
  if(sign_io != sign_input2)
  {
    if((sign_io != 0) && (!(io[0] & sign_bit)))
    {
      /* First input was negative and is now positive
         --> underflow */
      mInt_SetLastError(env,this,mInt_underflow);
      ret_val = JNI_FALSE;
    }
    else
      if((sign_io == 0) && (io[0] & sign_bit))
      {
        /* First input was positive and is now negative
           --> over */
        mInt_SetLastError(env,this,mInt_overflow);
        ret_val = JNI_FALSE;
      }
  }

  (*env)->ReleaseIntArrayElements(env,input2_array,input2,0);
  (*env)->ReleaseIntArrayElements(env,io_array,io,0);

  return ret_val;
}

/* Multiplication */

JNIEXPORT jboolean JNICALL Java_mInt_Multiply(JNIEnv *env,
  jobject this, jobject other)
{
  int io_sign;
  int input2_sign;
  jboolean ret_val;

  jintArray io_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *io = (*env)->GetIntArrayElements(env,io_array,0);

  jintArray input2_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *input2 = (*env)->GetIntArrayElements(env,input2_array,0);

  io_sign = io[0] & sign_bit;
  input2_sign = input2[0] & sign_bit;

  /* Use Negate since we can't *flow going from neg to pos */
  if(io_sign)
    Negate(io);

  if(input2_sign)
    Negate(input2);

  ret_val = UMultiply(env,this,io,input2);

  /* Restore input2's sign, as needed */
  if(input2_sign)
    Negate(input2);

  if(ret_val)
  {
    if(io_sign != input2_sign)
    {
      /* Input signs differed --> make output neg
         and check for underflow */
      if(!PosToNeg(env,this,io))
      {
        mInt_SetLastError(env,this,mInt_underflow);
        ret_val = JNI_FALSE;
      }
    }
    else
      if(io[0] & sign_bit)
      {
        /* Inputs were the same sign, output stays
           positive, but if MSb is on, it overflowed */
        mInt_SetLastError(env,this,mInt_overflow);
        ret_val = JNI_FALSE;
      }
  }

  (*env)->ReleaseIntArrayElements(env,input2_array,input2,0);
  (*env)->ReleaseIntArrayElements(env,io_array,io,0);

  return ret_val;
}

/* Division */

JNIEXPORT jboolean JNICALL Java_mInt_Divide(JNIEnv *env,
  jobject this, jobject other)
{
  jboolean ret_val;

  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);

  jintArray y_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *y = (*env)->GetIntArrayElements(env,y_array,0);

  int x_sign = x[0] & sign_bit;
  int y_sign = y[0] & sign_bit;
  int remainder[ints_in_mint];    /* Local temp */

  /* Use Negate since we can't *flow going from neg to pos */
  if(x_sign)
    Negate(x);

  if(y_sign)
    Negate(y);

  ret_val = UDivide(env,this,x,y,x,remainder);

  /* If signs differed, make quotient neg */
  if(x_sign != y_sign)
    Negate(x);

  (*env)->ReleaseIntArrayElements(env,y_array,y,0);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);

  return ret_val;
}

/* Remainder from division */

JNIEXPORT jboolean JNICALL Java_mInt_Remainder(JNIEnv *env,
  jobject this, jobject other)
{
  jboolean ret_val;

  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);

  jintArray y_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *y = (*env)->GetIntArrayElements(env,y_array,0);

  int x_sign = x[0] & sign_bit;
  int quotient[ints_in_mint];

  /* Use Negate since we can't *flow going from neg to pos */
  if(x_sign)
    Negate(x);

  if(y[0] & sign_bit)
    Negate(y);

  ret_val = UDivide(env,this,x,y,quotient,x);

  /* Remainder's sign is the same as the numerator's */
  if(x_sign)
    Negate(x);

  (*env)->ReleaseIntArrayElements(env,y_array,y,0);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);

  return ret_val;
}

/* Compare two mInts, return -1, 0, 1 if this is <, ==, or > x */

JNIEXPORT jint JNICALL Java_mInt_Comp(JNIEnv *env, jobject this,
  jobject other)
{
  jint ret_val;

  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);

  jintArray y_array = (*env)->GetObjectField(env,other,
    global_int_FID);
  jint *y = (*env)->GetIntArrayElements(env,y_array,0);

  int x_sign = x[0] & sign_bit;
  int y_sign = y[0] & sign_bit;

  /* If signs differ we can use them to decide the issue
     without examining the rest of the numbers */
  if(x_sign != y_sign)
  {
    /* x is pos, y is neg --> x > y */
    if(x_sign == 0)
      ret_val = 1;
    /* x is neg, y is pos --> x < y */
    else
      ret_val = -1;
  }
  else
    ret_val = UComp(x,y);

  (*env)->ReleaseIntArrayElements(env,y_array,y,0);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);

  return ret_val;
}

/* Negate the input via two's complement */

JNIEXPORT void JNICALL Java_mInt_Negate(JNIEnv *env, jobject this)
{
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);
  Negate(x);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);
}

/* ShiftLeft: Shift the mInt left 1 bit */

JNIEXPORT jboolean JNICALL Java_mInt_ShiftLeft(JNIEnv *env, jobject this)
{
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);
  jboolean ret_val = ShiftLeft(env,this,x);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);
  return ret_val;
}

/* ShiftRight: Shift the mInt right 1 bit */

JNIEXPORT void JNICALL Java_mInt_ShiftRight(JNIEnv *env, jobject this)
{
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);
  ShiftRight(x);
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);
}

/* Local function: Set last_error */

static void mInt_SetLastError(JNIEnv *env, jobject this,
  int new_last_error)
{
  jclass the_class = (*env)->GetObjectClass(env,this);
  (*env)->SetStaticIntField(env,the_class,global_le_FID,
    new_last_error);
}

/* Local function: Negate the input via two's complement */

void Negate(jint *x)
{
__asm
  {
    mov edi,x

    not dword ptr[edi+28]
    not dword ptr[edi+24]
    not dword ptr[edi+20]
    not dword ptr[edi+16]
    not dword ptr[edi+12]
    not dword ptr[edi+8]
    not dword ptr[edi+4]
    not dword ptr[edi]

    mov eax,1
    add [edi+28],eax

    mov eax,0
    adc [edi+24],eax
    adc [edi+20],eax
    adc [edi+16],eax
    adc [edi+12],eax
    adc [edi+8],eax
    adc [edi+4],eax
    adc [edi],eax
  }
}

/* Local function: Unsigned addition */

jboolean UAdd(JNIEnv *env, jobject this, jint *io,
  jint *input2)
{
  int temp = 0;

__asm
  {
    mov edi,io
    mov esi,input2

    mov eax,[esi+28]
    add [edi+28],eax

    mov eax,[esi+24]
    adc [edi+24],eax

    mov eax,[esi+20]
    adc [edi+20],eax

    mov eax,[esi+16]
    adc [edi+16],eax

    mov eax,[esi+12]
    adc [edi+12],eax

    mov eax,[esi+8]
    adc [edi+8],eax

    mov eax,[esi+4]
    adc [edi+4],eax

    mov eax,[esi]
    adc [edi],eax

    jnc no_carry

    mov eax,1
    mov temp,eax

  no_carry:
  }

  if(temp != 0)
  {
    mInt_SetLastError(env,this,mInt_overflow);
    return JNI_FALSE;
  }

  return JNI_TRUE;
}

/* Local function: Unsigned subtraction */

jboolean USubtract(JNIEnv *env, jobject this, jint *io,
  jint *input2)
{
  int temp = 0;

__asm
  {
    mov edi,io
    mov esi,input2

    mov eax,[esi+28]
    sub [edi+28],eax

    mov eax,[esi+24]
    sbb [edi+24],eax

    mov eax,[esi+20]
    sbb [edi+20],eax

    mov eax,[esi+16]
    sbb [edi+16],eax

    mov eax,[esi+12]
    sbb [edi+12],eax

    mov eax,[esi+8]
    sbb [edi+8],eax

    mov eax,[esi+4]
    sbb [edi+4],eax

    mov eax,[esi]
    sbb [edi],eax

    jnc no_carry

    mov eax,1
    mov temp,eax

  no_carry:
  }

  if(temp != 0)
  {
    mInt_SetLastError(env,this,mInt_underflow);
    return JNI_FALSE;
  }

  return JNI_TRUE;
}

/* Local function: Unsigned multiplication */

jboolean UMultiply(JNIEnv *env, jobject this, jint *io,
  jint *input2)
{
  jint result[ints_in_mint];
  int shift_count, bits, i;

  UInt32ToInt(result,0);
  shift_count = LeadingZeroes(io);
  ShiftLeftNNoError(io,shift_count);
  bits = bits_in_mint - shift_count;

  for(i = 0; i < bits; i++)
  {
    /* Only shift left on iterations 1..bits, and terminate
       if the shift results in a shift out of the result */
    if(i > 0)
      if(!ShiftLeft(env,this,result))
        return JNI_FALSE;

    /* If the next bit is on, we add in the other input */
    if(io[0] & sign_bit)
      if(!UAdd(env,this,result,input2))
        return JNI_FALSE;

    /* Always shift to the next bit in the first input */
    ShiftLeftNoError(io);
  }

  /* We got here, all is OK, so copy the result to the
     first parameter and exit OK. */
  memcpy(io,result,bytes_in_mint);

  return JNI_TRUE;
}

/* Local function: Unsigned division */

jboolean UDivide(JNIEnv *env, jobject this, jint *numer,
  jint *denom, jint *quotient, jint *remainder)
{
  int i, compare, shift_count = 0;
  int local_denom[ints_in_mint], local_quotient[ints_in_mint],
    local_rem[ints_in_mint];

  /* Special case: denominator is 0 */
  if(IsZero(denom))
  {
    for(i = 1; i < ints_in_mint; i++)
      quotient[i] = all_ones;

    quotient[0] = max_int;

    UInt32ToInt(remainder,0);

    mInt_SetLastError(env,this,mInt_divide_by_zero);
    return JNI_FALSE;
  }

  /* Special case: numerator is 0 */
  if(IsZero(numer))
  {
    UInt32ToInt(quotient,0);
    UInt32ToInt(remainder,0);
    return JNI_TRUE;
  }

  compare = UComp(numer,denom);

  /* Special case: denominator > numerator */
  if(compare < 0)
  {
    memcpy(remainder,numer,bytes_in_mint);
    UInt32ToInt(quotient,0);
    return JNI_TRUE;
  }

  /* Special case: denominator == numerator */
  if(compare == 0)
  {
    UInt32ToInt(quotient,1);
    UInt32ToInt(remainder,0);
	return JNI_TRUE;
  }

  /* We have to do a real division operation.  Start by
     shifting the denom. left until it and the numerator
     have the same number of leading zeroes.  This ensures
     that the denom. is as large as we can make it without
     exceeding the numerator's value, because of the special
     cases we've already handled above. */

  memcpy(local_denom,denom,bytes_in_mint);
  memcpy(local_rem,numer,bytes_in_mint);
  UInt32ToInt(local_quotient,0);

  /* The next two lines greatly increase the performance of
     the entire function */
  shift_count = LeadingZeroes(local_denom) - LeadingZeroes(local_rem) + 1;
  ShiftLeftNNoError(local_denom,shift_count - 1);

  for(i = 0; i < shift_count; i++)
  {
    ShiftLeftNoError(local_quotient);

    if(UComp(local_rem,local_denom) >= 0 )
    {
       USubtract(env,this,local_rem,local_denom);
       local_quotient[ints_in_mint - 1] += 1;
    }

    ShiftRight(local_denom);
  }

  memcpy(remainder,local_rem,bytes_in_mint);
  memcpy(quotient,local_quotient,bytes_in_mint);

  return JNI_TRUE;
}

/* Local function: Unsigned comparison */

jint UComp(jint *x, jint *y)
{
  int i;

  for(i = 0; i < ints_in_mint; i++)
  {
    if((unsigned)x[i] > (unsigned)y[i])
      return 1;

    if((unsigned)x[i] < (unsigned)y[i])
      return -1;
  }

  return 0;
}

/* Local function: Convert unsigned 32-bit integer to a mInt */

void UInt32ToInt(jint *x, unsigned int y)
{
  int i;
  x[ints_in_mint - 1] = y;

  for(i = 0; i < ints_in_mint - 1; i++)
    x[i] = 0;
}

/* Local function: Count the leading binary zeroes in a mInt */

int LeadingZeroes(jint *x)
{
  int i, zeroes = 0, temp;

  for(i = 0; i < ints_in_mint; i++)
  {
    temp = lz_tab[HIBYTE(HIWORD(x[i]))];
    zeroes += temp;

    if(temp < 8)
      return zeroes;

    temp = lz_tab[LOBYTE(HIWORD(x[i]))];
    zeroes += temp;

    if(temp < 8)
      return zeroes;

    temp = lz_tab[HIBYTE(LOWORD(x[i]))];
    zeroes += temp;

    if(temp < 8)
      return zeroes;

    temp = lz_tab[LOBYTE(LOWORD(x[i]))];
    zeroes += temp;

    if(temp < 8)
      return zeroes;
  }

  return zeroes;
}

/* Local function: Shift left with shift-out detection */

jboolean ShiftLeft(JNIEnv *env, jobject this, jint *x)
{
  int temp = 0;

__asm
  {
    mov esi,x

    shl DWORD PTR [esi+28],1
    rcl DWORD PTR [esi+24],1
    rcl DWORD PTR [esi+20],1
    rcl DWORD PTR [esi+16],1
    rcl DWORD PTR [esi+12],1
    rcl DWORD PTR [esi+8],1
    rcl DWORD PTR [esi+4],1
    rcl DWORD PTR [esi],1

    jnc no_carry

    mov eax,1
    mov temp,1

  no_carry:
  }

  if(temp != 0)
  {
    mInt_SetLastError(env,this,mInt_overflow);
    return JNI_FALSE;
  }

  return JNI_TRUE;
}

/* Local function: Shift right w/o shift out detection */

void ShiftRight(jint *x)
{
__asm
  {
    mov esi,x

    shr DWORD PTR [esi],1
    rcr DWORD PTR [esi+4],1
    rcr DWORD PTR [esi+8],1
    rcr DWORD PTR [esi+12],1
    rcr DWORD PTR [esi+16],1
    rcr DWORD PTR [esi+20],1
    rcr DWORD PTR [esi+24],1
    rcr DWORD PTR [esi+28],1
  }
}

/* Local function: Shift left w/o shift out detection */

void ShiftLeftNoError(jint *x)
{
__asm
  {
    mov esi,x

    shl DWORD PTR [esi+28],1
    rcl DWORD PTR [esi+24],1
    rcl DWORD PTR [esi+20],1
    rcl DWORD PTR [esi+16],1
    rcl DWORD PTR [esi+12],1
    rcl DWORD PTR [esi+8],1
    rcl DWORD PTR [esi+4],1
    rcl DWORD PTR [esi],1
  }
}

/* Local function: Unsigned shift left N bits w/o shift
   out detection */

void ShiftLeftNNoError(jint *x, int shift_bits)
{
  int temp, mask_shift, carry, new_val, i;
  unsigned int mask;

  /* Need one mask for each bit in a component */
  unsigned int msb_mask[32] =
    { 0x00000000, 0x80000000, 0xc0000000, 0xe0000000,
      0xf0000000, 0xf8000000, 0xfc000000, 0xfe000000,
      0xff000000, 0xff800000, 0xffc00000, 0xffe00000,
      0xfff00000, 0xfff80000, 0xfffc0000, 0xfffe0000,
      0xffff0000, 0xffff8000, 0xffffc000, 0xffffe000,
      0xfffff000, 0xfffff800, 0xfffffc00, 0xfffffe00,
      0xffffff00, 0xffffff80, 0xffffffc0, 0xffffffe0,
      0xfffffff0, 0xfffffff8, 0xfffffffc, 0xfffffffe };

  if(shift_bits == 0)
    return;

  if(shift_bits == 1)
  {
    ShiftLeftNoError(x);
    return;
  }

  while(shift_bits >= 32)
  {
    ShiftLeft32NoError(x);
    shift_bits -= 32;
  }

  if(shift_bits == 0)
    return;

  temp = 0;
  mask_shift = bits_in_component - shift_bits;
  carry = 0;
  mask = msb_mask[shift_bits];

  for(i = ints_in_mint - 1; i >= 0; i--)
  {
    new_val = carry + (x[i] << shift_bits);
    carry = (x[i] & mask) >> mask_shift;
    x[i] = new_val;
  }
}

/* Local function: Shift left 32 bits w/o shift out
   detection */

void ShiftLeft32NoError(jint *x)
{
  int i;

  for(i = 0; i < ints_in_mint - 1; i++)
    x[i] = x[i + 1];

  x[ints_in_mint - 1] = 0;
}

/* Local function: Flip a mint from pos to neg */

jboolean PosToNeg(JNIEnv *env, jobject this, jint *x)
{
  BOOL ret_val = JNI_TRUE;

  Negate(x);

  if((x[0] & sign_bit) || IsZero(x))
    return JNI_TRUE;
  else
  {
    mInt_SetLastError(env,this,mInt_underflow);
    return JNI_FALSE;
  }
}

/* Local function: Is mInt zero? */

BOOL IsZero(jint *x)
{
  int i;

  for(i = 0; i < ints_in_mint; i++)
    if(x[i] != 0)
      return FALSE;

  return TRUE;
}

/* HexString: Convert a mInt to a ASCII representation of
   hex bytes */

JNIEXPORT jstring JNICALL Java_mInt_HexString(JNIEnv *env,
  jobject this)
{
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);
  int i;

  char z[13];    /* 12 chars plus a null per int32 */

  /* Account for the hex chars plus the needed UTF-8-style
     double null */

  char output [max_hex_chars_in_mint + 2];

  output[0] = '\0';

  for(i = 0; i < ints_in_mint; i++)
  {
    sprintf(z,"%02hx %02hx %02hx %02hx ",
      HIBYTE(HIWORD(x[i])),LOBYTE(HIWORD(x[i])),
      HIBYTE(LOWORD(x[i])),LOBYTE(LOWORD(x[i])));

    strcat(output,z);
  }

  (*env)->ReleaseIntArrayElements(env,x_array,x,0);

  /* Double the null the make NewStringUTF happy, and also
     overlay the extraneous trailing blank */
  output [max_hex_chars_in_mint] = '\0';

  return (*env)->NewStringUTF(env,output);
}

/* toString: Convert a mInt to a printable String */

JNIEXPORT jstring JNICALL Java_mInt_toString(JNIEnv *env,
  jobject this)
{
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);
  int digit_counter;
  jint x_copy[ints_in_mint];
  jint rem[ints_in_mint];

  /* Account for the dec chars plus the needed UTF-8-style
     double null */

  char output [max_dec_chars_in_mint + 2];
  char *output_ptr = output;
  char digit_stack[max_dec_chars_in_mint + 1];

  output[0] = '\0';

  /* Use a local copy so we can change it in place */
  memcpy(x_copy,x,bytes_in_mint);

  /* If the mint is negative emit the leading '-' and negate
     the value */
  if(x_copy[0] & sign_bit)
  {
    Negate(x_copy);
    strcpy(output,"-");
    output_ptr++;
  }

  /* Special case: 0 */
  if(IsZero(x_copy))
  {
    strcpy(output,"0");
    goto done;
  }

  digit_counter = 0;

  while(!IsZero(x_copy))
  {
    /* Divide the input by 10 */

	UDivide(env,this,x_copy,mint_10,x_copy,rem);

    /* Pick up the remainder from input and convert to
       ASCII on the fly (We can get away with this hacked
       conversion because we know the remainder has to be
       in the range 0..9.) */
    digit_stack[digit_counter] =
      LOBYTE(LOWORD(rem[ints_in_mint - 1])) + 0x30;
    digit_counter++;
  }

  /* Copy the ASCII digits to the output string. */
  while(digit_counter > 0)
  {
    digit_counter--;
    *output_ptr = digit_stack[digit_counter];
    output_ptr++;
  }

  /* Null-terminate the output */
  *output_ptr = '\0';
  *(output_ptr + 1) = '\0';

done:
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);
  return (*env)->NewStringUTF(env,output);
}

/* SetFromString: ASCII to mInt conversion */

JNIEXPORT jint JNICALL Java_mInt_SetFromString(JNIEnv *env,
  jobject this, jstring js)
{
  const char *s = (*env)->GetStringUTFChars(env,js,0);
  jboolean ret_val = JNI_TRUE;
  jintArray x_array = (*env)->GetObjectField(env,this,
    global_int_FID);
  jint *x = (*env)->GetIntArrayElements(env,x_array,0);
  unsigned int i;
  BOOL was_neg = JNI_FALSE;
  jint temp[ints_in_mint] = { 0, 0, 0, 0, 0, 0, 0, 0 };

  UInt32ToInt(x,0);

  if(s[0] == '-')
  {
    was_neg = TRUE;
    s++;
  }

  for(i = 0; i < strlen(s); i++)
  {
    if(s[i] != ',')
    {
      if(!isdigit(s[i]))
      {
        mInt_SetLastError(env,this,mInt_invalid_char);
        ret_val = JNI_FALSE;
        goto done;
      }

      if(i > 0)
        if(!UMultiply(env,this,x,mint_10))
        {
          ret_val = JNI_FALSE;
          goto done;
        }

      temp[ints_in_mint - 1] = (int)(s[i] ^ 0x30);

      if(!UAdd(env,this,x,temp))
        {
          ret_val = JNI_FALSE;
          goto done;
        }
    }
  }

  if(was_neg)
    if(!PosToNeg(env,this,x))
      ret_val = JNI_FALSE;

done:
  (*env)->ReleaseIntArrayElements(env,x_array,x,0);
  (*env)->ReleaseStringUTFChars(env,js,s);

  return ret_val;
}

